@@ -14,7 +14,10 @@
  * See the License for the specific language governing permissions and
  * limitations under the License.
  */
-#include "mir_test/fake_event_hub.h"
+#include "mir/test/fake_event_hub.h"
+
+#include "mir/log.h"
+#include "mir/logging/logger.h"
 
 // from android-input
 #include <androidfw/Keyboard.h>
@@ -23,6 +26,8 @@
 #include <thread>
 #include <chrono>
 
+#include <sys/eventfd.h>
+
 using droidinput::AxisInfo;
 using droidinput::InputDeviceIdentifier;
 using droidinput::PropertyMap;
@@ -51,6 +56,7 @@ int const FakeEventHub::TouchScreenMinAxisValue = 0;
 int const FakeEventHub::TouchScreenMaxAxisValue = 100;
 
 FakeEventHub::FakeEventHub()
+    : trigger_fd{eventfd(0, EFD_CLOEXEC|EFD_NONBLOCK)}
 {
     keymap.loadGenericMaps();
 }
@@ -219,12 +225,14 @@ void FakeEventHub::setExcludedDevices(const Vector<String8>& devices)
     excluded_devices = devices;
 }
 
-size_t FakeEventHub::getEvents(int timeoutMillis, RawEvent* buffer, size_t bufferSize)
+size_t FakeEventHub::getEvents(RawEvent* buffer, size_t bufferSize)
 {
     size_t num_events_obtained = 0;
-    (void) timeoutMillis;
     {
         std::lock_guard<std::mutex> lg(guard);
+        uint64_t dummy;
+        if (sizeof dummy != read(trigger_fd, &dummy, sizeof dummy))
+            mir::log(mir::logging::Severity::debug, "FakeEventHub", "No event trigger to consume");
 
         if (throw_in_get_events)
             throw std::runtime_error("FakeEventHub::getEvents() exception");
@@ -404,8 +412,24 @@ void FakeEventHub::requestReopenDevices()
 {
 }
 
+void FakeEventHub::wakeIn(int32_t)
+{
+}
+
 void FakeEventHub::wake()
 {
+    uint64_t one{1};
+    if (sizeof one != write(trigger_fd, &one, sizeof one))
+        BOOST_THROW_EXCEPTION((std::system_error{errno,
+                               std::system_category(),
+                               "Failed to wake fake event hub"}));
+}
+
+void FakeEventHub::wake(droidinput::RawEvent const& event)
+{
+    std::lock_guard<std::mutex> lg(guard);
+    events_available.push_back(event);
+    wake();
 }
 
 void FakeEventHub::dump(droidinput::String8& dump)
@@ -421,6 +445,11 @@ void FakeEventHub::flush()
 {
 }
 
+mir::Fd FakeEventHub::fd()
+{
+    return trigger_fd;
+}
+
 void FakeEventHub::synthesize_builtin_keyboard_added()
 {
     RawEvent event;
@@ -428,8 +457,7 @@ void FakeEventHub::synthesize_builtin_keyboard_added()
     event.deviceId = BuiltInKeyboardID;
     event.type = EventHubInterface::DEVICE_ADDED;
 
-    std::lock_guard<std::mutex> lg(guard);
-    events_available.push_back(event);
+    wake(event);
 }
 
 void FakeEventHub::synthesize_builtin_cursor_added()
@@ -439,8 +467,7 @@ void FakeEventHub::synthesize_builtin_cursor_added()
     event.deviceId = BuiltInCursorID;
     event.type = EventHubInterface::DEVICE_ADDED;
 
-    std::lock_guard<std::mutex> lg(guard);
-    events_available.push_back(event);
+    wake(event);
 }
 
 void FakeEventHub::synthesize_usb_touchscreen_added()
@@ -469,6 +496,7 @@ void FakeEventHub::synthesize_usb_touchscreen_added()
     
     std::lock_guard<std::mutex> lg(guard);
     events_available.push_back(event);
+    wake();
 }
 
 void FakeEventHub::synthesize_device_scan_complete()
@@ -477,8 +505,7 @@ void FakeEventHub::synthesize_device_scan_complete()
     event.when = std::chrono::nanoseconds(0);
     event.type = EventHubInterface::FINISHED_DEVICE_SCAN;
 
-    std::lock_guard<std::mutex> lg(guard);
-    events_available.push_back(event);
+    wake(event);
 }
 
 void FakeEventHub::synthesize_event(const mis::KeyParameters &parameters)
@@ -498,8 +525,7 @@ void FakeEventHub::synthesize_event(const mis::KeyParameters &parameters)
     else
         event.value = 0;
 
-    std::lock_guard<std::mutex> lg(guard);
-    events_available.push_back(event);
+    wake(event);
 }
 
 void FakeEventHub::synthesize_event(const mis::ButtonParameters &parameters)
@@ -526,6 +552,7 @@ void FakeEventHub::synthesize_event(const mis::ButtonParameters &parameters)
     event.type = EV_SYN;
     event.code = SYN_REPORT;
     events_available.push_back(event);
+    wake();
 }
 
 void FakeEventHub::synthesize_event(const mis::MotionParameters &parameters)
@@ -551,6 +578,7 @@ void FakeEventHub::synthesize_event(const mis::MotionParameters &parameters)
     event.type = EV_SYN;
     event.code = SYN_REPORT;
     events_available.push_back(event);
+    wake();
 }
 
 void FakeEventHub::synthesize_event(const mis::TouchParameters &parameters)
@@ -584,6 +612,7 @@ void FakeEventHub::synthesize_event(const mis::TouchParameters &parameters)
     event.type = EV_SYN;
     event.code = SYN_REPORT;
     events_available.push_back(event);
+    wake();
 }
 
 void FakeEventHub::synthesize_event(std::chrono::nanoseconds when, int32_t device_id, int32_t type, int32_t code, int32_t value)
@@ -595,10 +624,7 @@ void FakeEventHub::synthesize_event(std::chrono::nanoseconds when, int32_t devic
     event.code = code;
     event.value = value;
 
-    {
-        std::lock_guard<std::mutex> lg(guard);
-        events_available.push_back(event);
-    }
+    wake(event);
 
     if (type == EV_ABS)
     {
@@ -763,4 +789,5 @@ void FakeEventHub::throw_exception_in_next_get_events()
 {
     std::lock_guard<std::mutex> lg(guard);
     throw_in_get_events = true;
+    wake();
 }
